import os
import traceback
from loguru import logger as log
from pandas import DataFrame
import numpy as np
from dycifer.utils import plotPrettyFFT
from dycifer.read import readSignals
from modelling_utils import stof, timer
from enum import Enum


def analogDynamicEval(subparser, *args, **kwargs):
    """_summary_
    Dynamic performance evaluation of Analog integrated circuits
    """
    import warnings

    warnings.filterwarnings("ignore")  # supress pandas warnings
    """_summary_
    Dynamic performance evaluation of Mixed Signals circuits
    """
    argv = None
    # read data
    sysargs = args[0]
    try:
        argv = subparser.parse_args(sysargs[1:])
    except Exception as e:
        log.error(traceback.format_exc())
    # from the signals argument (containing the signals file filepath)
    # extract the signals
    signals = readSignals(argv.signals[0])
    if argv.continuous_aos:
        print(
            "Running Continuous Amplitude Output System (CAOS) Dynamic Performance Evaluation..."
        )
        sampling_freq = stof(
            argv.sampling_frequency[0]
        )  # convert the parsed string to a float
        input_signal = argv.input_signal[0] if bool(argv.input_signal) else None
        output_signal = argv.output_signal[0]
        harmonics = argv.harmonics[0] if bool(argv.harmonics) else 7
        signal_span = argv.signal_span[0] if bool(argv.signal_span) else 0.0
        noise_power = argv.noise_power[0] if bool(argv.noise_power) else -1.0
        # pdb.set_trace()
        # perform dynamic performance evaluation
        (
            spectrum,
            target_harmonics,
            signal_power,
            dc_power,
            gain,
            gain_db,
            sfdr,
            thd,
            snr,
            sndr,
            hd2,
            hd3,
        ) = caosDynamicEval(
            signals,
            sampling_freq,
            output_signal,
            input_signal_name=input_signal,
            harmonics=harmonics,
            signal_span_factor=signal_span,
            noise_power=noise_power,
        )
        # prepare to plot resulting information
        dynamic_eval_indicators = DataFrame(
            data={
                "Output Signal Power (dB)": signal_power,
                "Output DC Power (dB)": dc_power,
                "Gain (out/in)": gain,
                "Gain (dB)": gain_db,
                "SFDR (dB)": sfdr,
                "THD (dB)": thd,
                "SNR (dB)": snr,
                "SNDR (dB)": sndr,
                "HD2": hd2,
                "HD3": hd3,
            },
            index=["Dynamic Evaluation Indicators"],
        )
        if argv.plot:
            plotPrettyFFT(
                spectrum[
                    spectrum.index.values >= 0
                ].index,  # plot only positive frequencies spectrum
                spectrum[spectrum.index.values >= 0]["power_db"],
                title="Signal Spectrum (dB)",
                xlabel="Frequency (GHz)",
                ylabel="Power (dB)",
                show=True,
                xlog=False,
                target_harmonics=target_harmonics,
                plot_to_terminal=argv.plot_to_terminal,
                xscale="G",
            )
        if bool(argv.output_file):
            plotPrettyFFT(
                spectrum[
                    spectrum.index >= 0
                ].index,  # plot only positive frequencies spectrum
                spectrum[spectrum.index.values >= 0]["power_db"],
                title="Signal Spectrum (dB)",
                xlabel="Frequency (GHz)",
                ylabel="Power (dB)",
                show=False,
                file_path=argv.output_file[0] + ".png",
                target_harmonics=target_harmonics,
                xscale="G",
            )
            if argv.generate_table:
                tablename = argv.output_file[0]
                dynamic_eval_indicators.to_csv(tablename + ".csv")
                dynamic_eval_indicators.to_json(tablename + ".json")
                dynamic_eval_indicators.to_markdown(tablename + ".md")
                dynamic_eval_indicators.to_latex(tablename + ".tex")
        # print indicators to console
        print()
        print(dynamic_eval_indicators.T)
    elif argv.discrete_aos:
        print(
            "Running Discrete Amplitude Output System (DAOS) Dynamic Performance Evaluation..."
        )
        sampling_freq = stof(
            argv.sampling_frequency[0]
        )  # convert the parsed string to a float
        wave_type = argv.waveform[0] if bool(argv.waveform) else "default"
        upper_level = argv.risetime_level[0] if bool(argv.risetime_level) else 0.9
        input_signal = argv.input_signal[0] if bool(argv.input_signal) else None
        output_signal = argv.output_signal[0]
        harmonics = argv.harmonics[0] if bool(argv.harmonics) else 7
        signal_span = argv.signal_span[0] if bool(argv.signal_span) else 0.0
        noise_power = argv.noise_power[0] if bool(argv.noise_power) else -1.0
        (
            spectrum,
            target_harmonics,
            signal_power,
            dc_power,
            gain,
            gain_db,
            sfdr,
            thd,
            snr,
            sndr,
            hd2,
            hd3,
            rise_time,
            bandwidth,
        ) = daosDynamicEval(
            signals,
            sampling_freq,
            output_signal,
            input_signal_name=input_signal,
            harmonics=harmonics,
            signal_span_factor=signal_span,
            noise_power=noise_power,
            levels=(0.1, upper_level),
            wave_type=wave_type,
            show_rise_time_eval=argv.plot,
        )
        # prepare to plot resulting information
        dynamic_eval_indicators = DataFrame(
            data={
                "Output Signal Power (dB)": signal_power,
                "Output DC Power (dB)": dc_power,
                "Gain (out/in)": gain,
                "Gain (dB)": gain_db,
                "SFDR (dB)": sfdr,
                "THD (dB)": thd,
                "SNR (dB)": snr,
                "SNDR (dB)": sndr,
                "HD2": hd2,
                "HD3": hd3,
                f"Rise Time[{100*upper_level}%] (ns)": rise_time / 1e-9,
                "Bandwidth (GHz)": bandwidth / 1e9,
            },
            index=["Dynamic Evaluation Indicators"],
        )
        if argv.plot:
            plotPrettyFFT(
                spectrum[
                    spectrum.index.values >= 0
                ].index,  # plot only positive frequencies spectrum
                spectrum[spectrum.index.values >= 0]["power_db"],
                title="Signal Spectrum (dB)",
                xlabel="Frequency (GHz)",
                ylabel="Power (dB)",
                show=True,
                xlog=False,
                target_harmonics=target_harmonics,
                plot_to_terminal=argv.plot_to_terminal,
                xscale="G",
            )
        if bool(argv.output_file):
            plotPrettyFFT(
                spectrum[
                    spectrum.index >= 0
                ].index,  # plot only positive frequencies spectrum
                spectrum[spectrum.index.values >= 0]["power_db"],
                title="Signal Spectrum (dB)",
                xlabel="Frequency (GHz)",
                ylabel="Power (dB)",
                show=False,
                file_path=argv.output_file[0] + ".png",
                target_harmonics=target_harmonics,
                xscale="G",
            )
            if argv.generate_table:
                tablename = argv.output_file[0]
                dynamic_eval_indicators.to_csv(os.path.join(tablename, ".csv"))
                dynamic_eval_indicators.to_json(os.path.join(tablename, ".json"))
                dynamic_eval_indicators.to_markdown(os.path.join(tablename, ".md"))
                dynamic_eval_indicators.to_latex(os.path.join(tablename, ".tex"))
        # print indicators to console
        print()
        print(dynamic_eval_indicators.T)
    else:
        raise ValueError(
            "No analog system class was specified was specified. Missing: --caos or --daos"
        )
    print("\nPerformance evaluation finished.")
    return


@timer
def caosDynamicEval(
    signals: DataFrame,
    sampling_frequency: float,
    output_signal_name: str,
    input_signal_name: str = None,
    harmonics: int = 7,
    signal_span_factor: float = 0.0,
    noise_power: float = -1.0,
) -> tuple[DataFrame, float, float, float, float, float, float, float, float]:
    """_summary_
    Dynamic performance evaluation of Continuous Analog Output Systems (CAOS)
    Args:
        signals (DataFrame): The time series data with all the correspondant signals.
        sampling_frequency (float): The sampling frequency of the signals.
        harmonics (int, optional): The number of harmonics to be used in the CAOS. Defaults to 7.
        signal_span_factor (float, optional): The factor to be used to scale the signal span. Defaults to 0.0.
        noise_power (float, optional): The noise power to be used in the CAOS. Defaults to -1.0.
        output_signal_name (str): The name of the output signal.
        input_signal_name (str, optional): The name of the input signal. Defaults to None. In no input signal is provided,
        the Gain of the system will not be computed.
    Returns:
        tuple[DataFrame, float, float, float, float, float, float, float]: The CAOS performance evaluation results.
            DataFrame: The frequency spectrum of the CAOS output signal in volt, volt squared (power in watt) and decibels.
            float(1): Output Signal's power in decibels
            float(2): Output Signal's DC power (in dB)
            float(3): Gain in linear scale
            float(4): Gain in dB
            float(5): Spurious Free Dynamic Range (SFDR) metric
            float(6): Total Harmonic Distortion (THD) metric
            float(7): Signal to Noise Ratio (SNR) metric
            float(8): Signal to Noise & Distortion Ratio (SNNDR) metric
            float(9): Fractional Second-Harmonic Distortion (HD2) metric
            float(10): Fractional Third-Harmonic Distortion (HD3) metric
    """
    downsampling = 1
    ts = 1.0 / sampling_frequency
    fs = 1.0 / ts
    if not bool(signals.index.name):
        # in case there is no time axis frame
        # automatically generate one from the sampling frequency

        t = np.arange(0, len(signals) * ts, ts)
        signals.set_index(t, inplace=True)
    else:
        downsampling = int(
            round(
                (1.0 / sampling_frequency)
                / (signals.index.values[1] - signals.index.values[0])
            )
        )
        if downsampling < 1:
            raise ValueError(
                "Sampling time period must be equal or higher than the signals' time resolution."
            )
    if not (output_signal_name in signals.columns):
        raise ValueError(f"{output_signal_name} does not belong to the parsed signals.")
    # perform downsampling if so was chosen
    signals = signals[::downsampling]
    # ts = signals.index.values[1] - signals.index.values[0]
    # fs = 1.0 / ts
    n_samples = len(signals.index)
    if noise_power > 0:
        noise_watt = (10 ** (noise_power / 10)) * 1e-3
        signals[output_signal_name] = signals[output_signal_name] + np.random.normal(
            0, np.sqrt(noise_watt), size=n_samples
        )
    vout = np.abs(
        np.fft.fftshift(np.fft.fft(signals[output_signal_name].values) / n_samples)
    )  # [V]
    freq = np.fft.fftshift(np.fft.fftfreq(len(vout), ts))  # [Hz]
    power = (
        vout * vout
    )  # [V^2] - square the voltage spectrum to obtain the power spectrum
    power_db = 10 * np.log10(power)  # [dB] - convert the power spectrum to dB
    spectrum = DataFrame(
        index=freq, data={"vout": vout, "power": power, "power_db": power_db}
    )
    # positive frequencies spectrum
    pspectrum = spectrum[spectrum.index >= 0].copy()
    # ********************************************
    # Obtaining the ADC's output signal power
    # ********************************************
    # determine the span of the signal's spectrum to consider it's total dispersed power
    span = np.max([1, int(np.floor(signal_span_factor * len(pspectrum.index)))])
    # obtain the signal frequency bin
    signal_bin = pspectrum["power"][
        span:
    ].idxmax()  # don't count DC signal when searching for the signal bin
    # obtain the harmonics of the signal from the signal bin
    harmonic_bins = [
        pspectrum.index[
            np.abs(pspectrum.index - (mult * signal_bin))
            == np.min(np.abs(pspectrum.index - (mult * signal_bin)))
        ][0]
        for mult in range(1, harmonics + 1)
        if mult * signal_bin <= np.max(freq)
    ]
    # tones that surpass Fs are aliased back to [0, Fs/2] spectrum
    harmonic_bins = [
        pspectrum.index[
            np.abs(pspectrum.index - (fs - bin))
            == np.min(np.abs(pspectrum.index - (fs - bin)))
        ]
        if bin > fs / 2
        else bin
        for bin in harmonic_bins
    ]
    # indexes of the harmonic bins
    harmonic_bins_idxs = [
        pspectrum.index.get_loc(bin) for bin in harmonic_bins if bin in pspectrum.index
    ]
    harmonics_power = np.array(
        [
            np.sum(
                pspectrum["power"]
                .iloc[harmonic_bin_idx - span : harmonic_bin_idx + span]
                .values
            )
            for harmonic_bin_idx in harmonic_bins_idxs
        ]
    )
    # obtain the signal_power
    signal_power = harmonics_power[0]
    SIGNAL_POWER_DB = 10 * np.log10(signal_power)
    signal_dc_power = np.sum(pspectrum["power"].iloc[0 : 0 + span].values)
    DC_POWER_DB = 10 * np.log10(signal_dc_power)
    # ********************************************
    # Computing SFDR - Spurious Free Dynamic Range
    #  - Obtain the power of the
    #       strongest spurious component
    #       of the spectrum (excluding the
    #       DC component) and compute the SFDR.
    # ********************************************
    signal_bin_idx = harmonic_bins_idxs[0]  # get the index of the signal bin
    spurious_spectrum = pspectrum["power"].copy()
    # erase the signal bin from the spurious spectrum
    spurious_spectrum.iloc[signal_bin_idx - span : signal_bin_idx + span] = np.min(
        pspectrum["power"]
    )
    # erase the signal's DC component from the spurious spectrum
    spurious_spectrum.iloc[0 : 0 + span] = np.min(pspectrum["power"])
    # find the strongest spurious component
    spur_bin = spurious_spectrum.idxmax()
    spur_bin_idx = pspectrum.index.get_loc(spur_bin)
    # measure the power of the strongest spurious component
    spur_power = np.sum(
        spurious_spectrum.iloc[spur_bin_idx - span : spur_bin_idx + span].values
    )
    # compute the SFDR
    SFDR = 10 * np.log10(signal_power / spur_power)
    # ********************************************
    # Computing THD - Total Harmonic Distortion
    #  - Obtain the power of each harmonic component
    # ********************************************
    # compute the total power of the sum of the harmonics
    THD = np.nan
    SNR = np.nan
    SNDR = np.nan
    HD2 = np.nan
    HD3 = np.nan
    try:
        total_distortion_power = np.sum(harmonics_power[1:])
        THD = 10 * np.log10(total_distortion_power / harmonics_power[0])
        # ********************************************
        # Computing SNR - Signal to Noise Ratio
        #  - Obtain the noise power in the spectrum
        # ********************************************
        noise_power = (
            np.sum(pspectrum["power"].values)
            - signal_dc_power
            - signal_power
            - total_distortion_power
        )
        SNR = 10 * np.log10(signal_power / noise_power)
        # ********************************************
        # Computing SNDR - Signal to Noise & Distortion Ratio
        #  - Add the noise and distortion power in
        #     the spectrum and compare them to the
        #     signal power
        # ********************************************
        SNDR = 10 * np.log10(signal_power / (noise_power + total_distortion_power))
        # ********************************************
        # Computing ENOB - Effective Number of Bits
        #  - Check deterioration of the ADC's
        #    ideal resolution because of the SNDR
        # ********************************************
        ENOB = (SNDR - 1.76) / 6.02
        # ********************************************
        # Computing HD2 and HD3 - Fractional Harmonic
        # Distortion of Second and Third order
        # harmonics
        # ********************************************
        HD2 = 10 * np.log10(harmonics_power[1] / harmonics_power[0])
        HD3 = 10 * np.log10(harmonics_power[2] / harmonics_power[0])
    except IndexError:
        log.warning(
            f"\nTried to access an harmonic that was not found in the frequency domain of the analysed spectrum.\nThis is likely due to the fact that the chosen sampling frequency is \nviolating the Nyquist Theorem in relation to the 2nd or 3rd Order Harmonics.\nTry to increase the sampling frequency."
        )
    # ********************************************
    # Computing Gain - Output Signal amplitude
    # to Input Signal amplitude ratio
    # ********************************************
    # measure the power of the fundamental harmonic of the input spectrum
    GAIN = np.nan
    GAIN_DB = np.nan
    if bool(input_signal_name):
        if not (input_signal_name in signals.columns):
            raise ValueError(
                f"{input_signal_name} does not belong to the parsed signals."
            )

        vin = np.abs(
            np.fft.fftshift(np.fft.fft(signals[input_signal_name].values) / n_samples)
        )
        freq_in = np.fft.fftshift(np.fft.fftfreq(len(vin), ts))  # [Hz]
        in_power = vin * vin
        in_power_db = np.nan_to_num(
            10 * np.log10(in_power), nan=0.0, posinf=0.0, neginf=0.0, copy=True
        )
        in_spectrum = DataFrame(
            index=freq_in,
            data={"vin": vin, "in_power": in_power, "in_power_db": in_power_db},
        )
        in_pspectrum = in_spectrum[in_spectrum.index >= 0].copy()
        in_signal_bin = in_pspectrum["in_power"][
            0 + span :
        ].idxmax()  # don't count DC signal when searching for the signal bin
        # obtain the harmonics of the signal from the signal bin
        in_harmonic_bins = [
            mult * in_signal_bin
            for mult in range(1, harmonics + 1)
            if mult * signal_bin <= np.max(freq)
        ]
        # tones that surpass Fs are aliased back to [0, Fs/2] spectrum
        in_harmonic_bins = [
            fs - bin if bin > fs / 2 else bin for bin in in_harmonic_bins
        ]
        # indexes of the harmonic bins
        in_harmonic_bins_idxs = [
            pspectrum.index.get_loc(bin) for bin in in_harmonic_bins
        ]
        input_signal_power = np.sum(
            in_pspectrum["in_power"]
            .iloc[in_harmonic_bins_idxs[0] : in_harmonic_bins_idxs[0] + span]
            .values
        )
        GAIN = np.sqrt(signal_power / input_signal_power)
        GAIN_DB = 20 * np.log10(GAIN)
    target_harmonics = list(zip(harmonic_bins, 10 * np.log10(harmonics_power)))
    return (
        spectrum,
        target_harmonics,
        SIGNAL_POWER_DB,
        DC_POWER_DB,
        GAIN,
        GAIN_DB,
        SFDR,
        THD,
        SNR,
        SNDR,
        HD2,
        HD3,
    )


class WaveTypes(Enum):
    """_summary_

    Args:
        PULSE (str): Pure Pulse waveform selector
        SAWTOOTH (str): Pure Sawtooth waveform selector
        DEFAULT (str): Default modulated, square waveform selector
    """

    PULSE = "pulse"
    SAWTOOTH = "sawtooth"
    DEFAULT = "default"


@timer
def daosDynamicEval(
    signals: DataFrame,
    sampling_frequency: float,
    output_signal_name: str,
    input_signal_name: str = None,
    harmonics: int = 7,
    signal_span_factor: float = 0.0,
    noise_power: float = -1.0,
    wave_type: str = "default",
    levels: tuple = (0.1, 0.9),
    show_rise_time_eval: bool = False,
) -> tuple[DataFrame, float, float, float, float, float, float, float, float, float]:
    from heapq import nlargest
    from warnings import warn
    from matplotlib.pyplot import (
        plot,
        show,
        xlabel,
        ylabel,
        grid,
        title,
        legend,
        arrow,
        rc,
        clf,
    )

    """_summary_
    Dynamic performance evaluation of Discrete Analog Output Systems (CAOS)
    Args:
        signals (DataFrame): The time series data with all the correspondant signals.
        sampling_frequency (float): The sampling frequency of the signals.
        output_signal_name (str): The name of the output signal.
        harmonics (int, optional): The number of harmonics to be used in the CAOS. Defaults to 7.
        signal_span_factor (float, optional): The factor to be used to scale the signal span. Defaults to 0.0.
        noise_power (float, optional): The noise power to be used in the CAOS. Defaults to -1.0.
        input_signal_name (str, optional): The name of the input signal. Defaults to None. In no input signal is provided,
        the Gain of the system will not be computed.
        wave_type (str, optional): The type of wave to be used in the DAOS. Defaults to "pulse". Options:
        levels (tuple, optional): The levels to be used in the Risetime computation. Defaults to (0.1, 0.9).
    Returns:
        tuple[DataFrame, float, float, float, float, float, float, float]: The CAOS performance evaluation results.
        DataFrame: The frequency spectrum of the CAOS output signal in volt, volt squared (power in watt) and decibels.
        float(1): Output Signal's power in decibels
        float(2): Output Signal's DC power (in dB)
        float(3): Gain in linear scale
        float(4): Gain in dB
        float(5): Spurious Free Dynamic Range (SFDR) metric
        float(6): Total Harmonic Distortion (THD) metric
        float(7): Signal to Noise Ratio (SNR) metric
        float(8): Signal to Noise & Distortion Ratio (SNNDR) metric
        float(9): Fractional Second-Harmonic Distortion (HD2) metric
        float(10): Fractional Third-Harmonic Distortion (HD3) metric
        float(11): Average Rise Time (ns) in 90% of the signal
        float(12): Estimated Bandwidth (Hz) of the output signal
    """
    downsampling = 1
    ts = 1.0 / sampling_frequency
    fs = 1.0 / ts
    if not bool(signals.index.name):
        # in case there is no time axis frame
        # automatically generate one from the sampling frequency
        # ts = 1.0 / sampling_frequency
        t = np.arange(0, len(signals) * ts, ts)
        signals.set_index(t, inplace=True)
    else:
        downsampling = int(
            round(
                (1.0 / sampling_frequency)
                / (signals.index.values[1] - signals.index.values[0])
            )
        )
        if downsampling < 1:
            raise ValueError(
                "Sampling time period must be equal or higher than the signals' time resolution."
            )
    if not (output_signal_name in signals.columns):
        raise ValueError(f"{output_signal_name} does not belong to the parsed signals.")
    # perform downsampling if so was chosen
    signals = signals[::downsampling]
    # ts = signals.index.values[1] - signals.index.values[0]
    # fs = 1.0 / ts
    n_samples = len(signals.index)
    if noise_power > 0:
        noise_watt = (10 ** (noise_power / 10)) * 1e-3
        signals[output_signal_name] = signals[output_signal_name] + np.random.normal(
            0, np.sqrt(noise_watt), size=n_samples
        )
    vout = np.abs(
        np.fft.fftshift(np.fft.fft(signals[output_signal_name].values) / n_samples)
    )  # [V]
    freq = np.fft.fftshift(np.fft.fftfreq(len(vout), ts))  # [Hz]
    power = (
        vout * vout
    )  # [V^2] - square the voltage spectrum to obtain the power spectrum
    power_db = 10 * np.log10(power)  # [dB] - convert the power spectrum to dB
    spectrum = DataFrame(
        index=freq, data={"vout": vout, "power": power, "power_db": power_db}
    )
    # positive frequencies spectrum
    pspectrum = spectrum[spectrum.index >= 0].copy()
    # ********************************************
    # Obtaining the ADC's output signal power
    # ********************************************
    # determine the span of the signal's spectrum to consider it's total dispersed power
    span = np.max([1, int(np.floor(signal_span_factor * len(pspectrum.index)))])
    # obtain the signal frequency bin
    signal_bin = pspectrum["power"][
        span:
    ].idxmax()  # don't count DC signal when searching for the signal bin
    # obtain the harmonics of the signal from the signal bin
    harmonic_bins = [
        pspectrum.index[
            np.abs(pspectrum.index - (mult * signal_bin))
            == np.min(np.abs(pspectrum.index - (mult * signal_bin)))
        ][0]
        for mult in range(1, harmonics + 1)
        if mult * signal_bin <= np.max(freq)
    ]
    # tones that surpass Fs are aliased back to [0, Fs/2] spectrum
    harmonic_bins = [
        pspectrum.index[
            np.abs(pspectrum.index - (fs - bin))
            == np.min(np.abs(pspectrum.index - (fs - bin)))
        ]
        if bin > fs / 2
        else bin
        for bin in harmonic_bins
    ]
    # indexes of the harmonic bins
    harmonic_bins_idxs = [pspectrum.index.get_loc(bin) for bin in harmonic_bins]
    harmonics_power = np.array(
        [
            np.sum(
                pspectrum["power"]
                .iloc[harmonic_bin_idx - span : harmonic_bin_idx + span]
                .values
            )
            for harmonic_bin_idx in harmonic_bins_idxs
        ]
    )
    # obtain the signal_power
    signal_power = harmonics_power[0]
    SIGNAL_POWER_DB = 10 * np.log10(signal_power)
    signal_dc_power = np.sum(pspectrum["power"].iloc[0 : 0 + span].values)
    DC_POWER_DB = 10 * np.log10(signal_dc_power)
    # ********************************************
    # Computing SFDR - Spurious Free Dynamic Range
    #  - Obtain the power of the
    #       strongest spurious component
    #       of the spectrum (excluding the
    #       DC component) and compute the SFDR.
    # ********************************************
    signal_bin_idx = harmonic_bins_idxs[0]  # get the index of the signal bin
    spurious_spectrum = pspectrum["power"].copy()
    # erase the signal bin from the spurious spectrum
    spurious_spectrum.iloc[signal_bin_idx - span : signal_bin_idx + span] = np.min(
        pspectrum["power"]
    )
    # erase the signal's DC component from the spurious spectrum
    spurious_spectrum.iloc[0 : 0 + span] = np.min(pspectrum["power"])
    # find the strongest spurious component
    spur_bin = spurious_spectrum.idxmax()
    spur_bin_idx = pspectrum.index.get_loc(spur_bin)
    # measure the power of the strongest spurious component
    spur_power = np.sum(
        spurious_spectrum.iloc[spur_bin_idx - span : spur_bin_idx + span].values
    )
    # compute the SFDR
    SFDR = 10 * np.log10(signal_power / spur_power)
    # ********************************************
    # Computing THD - Total Harmonic Distortion
    #  - Obtain the power of each harmonic component
    # ********************************************
    # compute the total power of the sum of the harmonics
    THD = np.nan
    SNR = np.nan
    SNDR = np.nan
    HD2 = np.nan
    HD3 = np.nan
    try:
        total_distortion_power = np.sum(harmonics_power[1:])
        THD = 10 * np.log10(total_distortion_power / harmonics_power[0])
        # ********************************************
        # Computing SNR - Signal to Noise Ratio
        #  - Obtain the noise power in the spectrum
        # ********************************************
        noise_power = (
            np.sum(pspectrum["power"].values)
            - signal_dc_power
            - signal_power
            - total_distortion_power
        )
        SNR = 10 * np.log10(signal_power / noise_power)
        # ********************************************
        # Computing SNDR - Signal to Noise & Distortion Ratio
        #  - Add the noise and distortion power in
        #     the spectrum and compare them to the
        #     signal power
        # ********************************************
        SNDR = 10 * np.log10(signal_power / (noise_power + total_distortion_power))
        # ********************************************
        # Computing ENOB - Effective Number of Bits
        #  - Check deterioration of the ADC's
        #    ideal resolution because of the SNDR
        # ********************************************
        ENOB = (SNDR - 1.76) / 6.02
        # ********************************************
        # Computing HD2 and HD3 - Fractional Harmonic
        # Distortion of Second and Third order
        # harmonics
        # ********************************************
        HD2 = 10 * np.log10(harmonics_power[1] / harmonics_power[0])
        HD3 = 10 * np.log10(harmonics_power[2] / harmonics_power[0])
    except IndexError:
        log.warning(
            f"\nTried to access an harmonic that was not found in the frequency domain of the analysed spectrum.\nThis is likely due to the fact that the chosen sampling frequency is \nviolating the Nyquist Theorem in relation to the 2nd or 3rd Order Harmonics.\nTry to increase the sampling frequency."
        )
    # ********************************************
    # Computing Gain - Output Signal amplitude
    # to Input Signal amplitude ratio
    # ********************************************
    # measure the power of the fundamental harmonic of the input spectrum and divide both output and input powers
    GAIN = np.nan
    GAIN_DB = np.nan
    if bool(input_signal_name):
        if not (input_signal_name in signals.columns):
            raise ValueError(
                f"{input_signal_name} does not belong to the parsed signals."
            )

        vin = np.abs(
            np.fft.fftshift(np.fft.fft(signals[input_signal_name].values) / n_samples)
        )
        freq_in = np.fft.fftshift(np.fft.fftfreq(len(vin), ts))  # [Hz]
        in_power = vin * vin
        in_power_db = np.nan_to_num(
            10 * np.log10(in_power), nan=0.0, posinf=0.0, neginf=0.0, copy=True
        )
        in_spectrum = DataFrame(
            index=freq_in,
            data={"vin": vin, "in_power": in_power, "in_power_db": in_power_db},
        )
        in_pspectrum = in_spectrum[in_spectrum.index >= 0].copy()
        in_signal_bin = in_pspectrum["in_power"][
            0 + span :
        ].idxmax()  # don't count DC signal when searching for the signal bin
        # obtain the harmonics of the signal from the signal bin
        in_harmonic_bins = [
            mult * in_signal_bin
            for mult in range(1, harmonics + 1)
            if mult * signal_bin <= np.max(freq)
        ]
        # tones that surpass Fs are aliased back to [0, Fs/2] spectrum
        in_harmonic_bins = [
            fs - bin if bin > fs / 2 else bin for bin in in_harmonic_bins
        ]
        # indexes of the harmonic bins
        in_harmonic_bins_idxs = [
            pspectrum.index.get_loc(bin) for bin in in_harmonic_bins
        ]
        input_signal_power = np.sum(
            in_pspectrum["in_power"]
            .iloc[in_harmonic_bins_idxs[0] : in_harmonic_bins_idxs[0] + span]
            .values
        )
        GAIN = np.sqrt(signal_power / input_signal_power)
        GAIN_DB = 20 * np.log10(GAIN)
    # ********************************************
    # Computing output signal's
    # risetime @90% signal variation
    # between static levels
    # ********************************************

    # ********************************************
    # Computing output signal's
    # falltime @90% signal variation
    # between static levels
    # ********************************************
    RISETIME_90 = 0.0
    vout = signals[output_signal_name].values
    if wave_type in [WaveTypes.SAWTOOTH.value, WaveTypes.PULSE.value]:
        threshold_vout = np.min(vout)
    elif wave_type is WaveTypes.DEFAULT.value:
        threshold_vout = np.mean(vout)
    else:
        raise ValueError(
            f"{wave_type} is not a valid wave type. Possible types are: {[elem.value for elem in WaveTypes]}."
        )
    # comput output signal histogram
    hist, bin_edges = np.histogram(vout)
    hist, bin_edges = hist / sum(hist), np.array(bin_edges)
    # through the histogram, find the discrete signal levels to be accounted for in the risetime
    hist_filtered = hist[bin_edges[:-1] > threshold_vout]
    signal_levels_prob = nlargest(2, hist_filtered)
    signal_levels = (
        bin_edges[:-1][bin_edges[:-1] > threshold_vout][
            hist_filtered == signal_levels_prob[0]
        ],
        bin_edges[:-1][bin_edges[:-1] > threshold_vout][
            hist_filtered == signal_levels_prob[1]
        ],
    )
    # upper_signal_levels = sorted((signal_levels[0][0], signal_levels[1][0]))
    signal_levels = signal_levels[0][0], signal_levels[1][0]
    s0, s100 = sorted(signal_levels)
    # measure all the transitions between these the signal levels found using the mean value of the signal
    delta_v = np.diff(vout)
    if levels[0] > levels[1]:
        raise ValueError(
            f"Rise Time computation levels must be in ascending order, but levels {levels} were given."
        )
    level0 = levels[0]
    level1 = levels[1]
    t10 = (s0 + (s100 - s0) * level0 - vout[:-1]) / delta_v
    idx10 = np.where((t10 > 0) & (t10 < 1))[0]
    t10 = idx10 + t10[idx10]
    t90 = (s0 + (s100 - s0) * level1 - vout[:-1]) / delta_v
    idx90 = np.where((t90 > 0) & (t90 < 1))[0]
    t90 = idx90 + t90[idx90]

    # compute all possible transition times, keep the smallest
    idx = np.unravel_index(np.argmin(np.abs(t90[:, None] - t10)), (len(t90), len(t10)))
    RISETIME_90 = (
        np.abs((t90[idx[0]] - t10[idx[1]])) * ts
    )  # finally, compute the risetime
    BANDWIDTH = 1 / RISETIME_90

    if show_rise_time_eval:
        log.warning(
            "\nRise Time Evaluation plotting is dehactivated due to some problems."
        )
        """
        rc("axes", titlesize=14)  # fontsize of the axes title
        rc("axes", labelsize=12)  # fontsize of the x and y labels
        rc("xtick", labelsize=12)  # fontsize of the tick labels
        rc("ytick", labelsize=12)  # fontsize of the tick labels
        rc("legend", fontsize=11)  # legend fontsize
        rc("font", size=11)  # controls default text sizes
        rc("font", family="serif")  # default serif font
        plot(signals.index, vout, label=output_signal_name)
        xlabel("Time [s]")
        ylabel("Voltage [V]")
        grid()
        title(
            f"Output signal RiseTime[{level0*100:.2f}% - {level1*100:.2f}%] Evaluation"
        )
        x = t10[idx[1]] * ts
        y = s0 + (s100 - s0) * level0
        dx = t90[idx[0]] * ts - x
        dy = s0 + (s100 - s0) * level1 - y
        # plot(t10*ts, [s0 + (s100-s0)*level0]*len(t10), 'go')
        plot(x, [y], "go")
        plot(
            x,
            [y],
            "o",
            mec="g",
            mfc="None",
            ms=10,
            label=f"Signal Level[{level0*100}%]",
        )
        # plot(t90*ts, [s0 + (s100-s0)*level1]*len(t90), 'ro')
        plot(x + dx, [y + dy], "ro")
        plot(
            x + dx,
            [y + dy],
            "o",
            mec="r",
            mfc="None",
            ms=10,
            label=f"Signal Level[{level1*100}%]",
        )
        # double sided arrow
        arrow(
            x,
            y,
            dx,
            dy,
            color="k",
            label=f"RiseTime @ [{level1*100}%] = {RISETIME_90/1e-9:.3f} ns",
            width=ts / 2,
            head_length=0.05,
            head_width=ts * 2,
            length_includes_head=True,
        )
        arrow(
            x + dx,
            y + dy,
            -dx,
            -dy,
            color="k",
            width=ts / 2,
            head_length=0.05,
            head_width=2 * ts,
            length_includes_head=True,
        )
        legend(loc="lower right")
        show()
        clf()
        """
    target_harmonics = list(zip(harmonic_bins, 10 * np.log10(harmonics_power)))
    return (
        spectrum,
        target_harmonics,
        SIGNAL_POWER_DB,
        DC_POWER_DB,
        GAIN,
        GAIN_DB,
        SFDR,
        THD,
        SNR,
        SNDR,
        HD2,
        HD3,
        RISETIME_90,
        BANDWIDTH,
    )



from loguru import logger as log
from dycifer import __version__, __author__, __email__
import sys
import os
import re
import argparse
import traceback
from dycifer.analog import analogDynamicEval
from dycifer.mixed_signals import mixedSignalsDynamicEval

# define the subparsers
__cmds = {
    "-ms": (
        "mixedsignals",
        "Mixed Signals dynamic performance evaluation",
        mixedSignalsDynamicEval,
    ),
    "-a": (
        "analog",
        "Analog integrated circuits dynamic performance evaluation",
        analogDynamicEval,
    ),
}
# define the arguments of each subparser
__cmd_args = {
    "-ms": {
        "-adc": (
            "--analog-to-digital",
            "Analog to Digital Converter (ADC) dynamic performance evaluation",
            "",
            bool,
            "opt",
        ),
        "-dac": (
            "--digital-to-analog",
            "Digital to Analog Converter (DAC) dynamic performance evaluation",
            "",
            bool,
            "opt",
        ),
        "-sda": (
            "--sigma-delta-adc",
            "Sigma Delta ADC dynamic performance evaluation",
            "",
            bool,
            "opt",
        ),
        "-sdd": (
            "--sigma-delta-dac",
            "Sigma Delta DAC dynamic performance evaluation",
            "",
            bool,
            "opt",
        ),
        "-fs": (
            "--sampling-frequency",
            "Sampling frequency of the parsed signals",
            "FREQUENCY",
            str,
            "",
        ),  # sampling frequency of the parsed signals is obligatory
        "-bit": (
            "--bit-resolution",
            "Bitwise resolution of the ADC/DAC",
            "#BITS",
            int,
            "opt",
        ),
        "-vs": (
            "--voltage-source",
            "Voltage source of the ADC/DAC",
            "VOLTAGE",
            float,
            "opt",
        ),
        "-nh": (
            "--harmonics",
            "Number of harmonics of the output signal power considered during THD computation",
            "#HARMONICS",
            int,
            "opt",
        ),
        "-ss": (
            "--signal-span",
            "Spectral dispertion factor of the output signal power",
            "DECIMAL",
            float,
            "opt",
        ),
        "-a": (
            "--ascending",
            "Signals of each Bit are in ascending bit order or not",
            "",
            bool,
            "opt",
        ),
    },
    "-a": {
        "-daos": (
            "--discrete-aos",
            "Discrete Amplitude Output system (DAOS) dynamic performance evaluation",
            "",
            bool,
            "opt",
        ),
        "-caos": (
            "--continuous-aos",
            "Continuous Amplitude Output system (CAOS) dynamic performance evaluation",
            "",
            bool,
            "opt",
        ),
        "-is": (
            "--input-signal",
            "Target output signal of the analysis",
            "NAME",
            str,
            "opt",
        ),
        "-fs": (
            "--sampling-frequency",
            "Sampling frequency of the parsed signals",
            "FREQUENCY",
            str,
            "",
        ),  # sampling frequency of the parsed signals is obligatory
        "-os": (
            "--output-signal",
            "Target output signal of the analysis",
            "NAME",
            str,
            "",
        ),
        "-wf": (
            "--waveform",
            'Target waveform type ("default": modulated pulse wave, "pulse": pulse wave, "sawtooth": sawtooth wave)',
            "TYPE",
            str,
            "opt",
        ),
        "-rl": (
            "--risetime-level",
            "Rise Time Computation Upper Level (usual: [0.9, 0.95, 0.99], default: 0.9)",
            "FLOAT[0;1]",
            float,
            "opt",
        ),
        "-nh": (
            "--harmonics",
            "Number of harmonics of the output signal power considered during THD computation",
            "#HARMONICS",
            int,
            "opt",
        ),
        "-ss": (
            "--signal-span",
            "Spectral dispertion factor of the output signal power",
            "DECIMAL",
            float,
            "opt",
        ),
    },
    "all": {
        "-s": (
            "--signals",
            "Path of the file containing the time-series signals to analyse",
            "FILEPATH",
            str,
            "",
        ),
        "-o": (
            "--output-file",
            "Specify the output filename/filepath for the resulting output files",
            "FILEPATH",
            str,
            "opt",
        ),
        "-np": (
            "--noise-power",
            "Power of the noise signal added to the target signals (in dBm)",
            "NOISE_DBM",
            float,
            "opt",
        ),
        "-p": (
            "--plot",
            "Boolean flag to indicate the programme to plot or not the resulting analysis graphs",
            "",
            bool,
            "opt",
        ),
        "-gt": (
            "--generate-table",
            "Boolean flag to indicate the programme to generate a markdown and a latex table with the resulting analysis data",
            "",
            bool,
            "opt",
        ),
        "-pt": (
            "--plot-to-terminal",
            "Plot the resulting analysis graphs to the terminal in ASCII mode",
            "",
            bool,
            "opt",
        ),
    },
}


def mapSubparserToFun(func, subparser):
    from functools import wraps

    """_summary_
    Maps subparser to callback function
    Args:
        func (Function): callback function
        subparser (_type_): subparser object
    Returns:
        result (_type_): result of the callback function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(subparser, *args, **kwargs)

    return wrapper


def setupParser(
    cmds: dict,
    args: dict,
    author: str = "",
    email: str = "",
    prog_name: str = "",
    prog_vers: str = "",
    desc: str = "",
    **kwargs,
) -> argparse.ArgumentParser:
    from colorama import init, Fore, Back, Style

    """_summary_
    Sets up the command line parser.
    Args:
        cmds (dict): dictionary of commands defining the callbacks to each subparser
        args (dict): dictionary of arguments for each command/subparser callback function
        entrymsg (str): entry message for the parser
        progname (str): program name
        progvers (str): program version
        desc (str): program description
    Returns:
        argparse.ArgumentParser: console argument parser object
    """
    if os.environ.get("OS", "") == "Windows_NT":
        init()  # init colorama
    gray, cyan, green, blue = Fore.WHITE, Fore.CYAN, Fore.GREEN, Fore.BLUE

    # setup the messages formatter class for the console
    col_base = gray if "col_base" not in kwargs else kwargs["col_base"]
    col_usage = cyan if "col_usage" not in kwargs else kwargs["col_usage"]
    col_option = cyan if "col_option" not in kwargs else kwargs["col_option"]
    col_caps = green if "col_caps" not in kwargs else kwargs["col_caps"]
    col_prog = green if "col_prog" not in kwargs else kwargs["col_prog"]

    def msg_formatter(msg, color=gray):
        return f"{color}{msg}{Style.RESET_ALL}"

    def entryMsg(prog_name: str = "", author: str = "", email: str = "") -> str:
        from pyfiglet import figlet_format

        """_summary_
        Returns the application entry message as a string.
        Returns:
            str: entry message
        """
        figlet_color = green
        figlet_art = (
            figlet_format(f"{prog_name.upper()}", font="alligator", justify="center")
            if bool(prog_name)
            else ""
        )
        msgs = [
            f"{msg_formatter(figlet_art, color=figlet_color)}",
            "by " + author + " (" + email + ")"
            if (bool(author) and bool(email))
            else "",
        ]
        max_ch = max([len(submsg) for submsg in [msgs[0][-1]] + msgs[1:]]) + 2
        sep_color = green
        sep = "#"
        # sep_msg = msg_formatter(sep * max_ch, color=sep_color)
        # ret = f"{sep_msg}" + "\n"
        ret = ""
        ret += msgs[0]
        for msg in msgs[1:]:
            ret += msg + " " * (max_ch - len(msg)) + "\n"
        # ret += f"{sep_msg}" + "\n"
        return ret

    def make_wide(formatter, w: int = 120, h: int = 36, max_help_position: int = 30):
        import warnings

        """Return a wider HelpFormatter, if possible."""
        try:
            # https://stackoverflow.com/a/5464440
            # beware: "Only the name of this class is considered a public API."
            kwargs = {
                "width": w,
                "max_help_position": h,
                "max_help_position": max_help_position,
            }
            formatter(None, **kwargs)
            return lambda prog: formatter(prog, **kwargs)
        except TypeError:
            warnings.warn("Argparse help formatter failed, falling back.")
            return formatter

    class FormattedConsoleParser(argparse.ArgumentParser):
        def _print_message(self, message, file=None):
            if message:
                if message.startswith("usage"):
                    print(entryMsg(prog_name, author, email))

                    message = f"{msg_formatter(prog_name, color=col_prog)} {prog_vers}\n\n{message}"
                    message = re.sub(
                        r"(-[a-z]+\s*|\[)([A-Z]+)(?=]|,|\s\s|\s\.)",
                        r"\1{}\2{}".format(col_caps, Style.RESET_ALL),
                        message,
                    )
                    message = re.sub(
                        r"((-|--)[a-z]+)",
                        r"{}\1{}".format(col_option, Style.RESET_ALL),
                        message,
                    )

                    usage = msg_formatter("USAGE", color=col_usage)
                    message = message.replace("usage", f"{usage}")

                    flags = msg_formatter("FLAGS", color=col_usage)
                    message = message.replace("options", f"{flags}")

                    progg = msg_formatter(self.prog, color=col_prog)
                    message = message.replace(self.prog, f"{progg}")
                message = f"{msg_formatter(message.strip(), color=col_base)}"
                print(message)

    parser = FormattedConsoleParser(
        prog=prog_name,
        description=desc,
        formatter_class=make_wide(argparse.HelpFormatter, max_help_position=45),
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"{msg_formatter(prog_name, color=col_prog)} v{msg_formatter(prog_vers, color=col_base)}",
    )
    # mutually exclusive arguments
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--quiet", action="store_true", help="quiet run")
    group.add_argument("--verbose", action="store_true", help="verbose run")
    subparsers = parser.add_subparsers()
    for cmd, (cmd_name, cmd_desc, cmd_callback) in cmds.items():
        subparser = subparsers.add_parser(
            cmd_name,
            help=cmd_desc,
            formatter_class=make_wide(argparse.HelpFormatter, max_help_position=45),
            fromfile_prefix_chars="@",  # to run the script from a text file
        )
        subparser.set_defaults(func=mapSubparserToFun(cmd_callback, subparser))
        optional = subparser._action_groups.pop()
        required = subparser.add_argument_group("required arguments")
        subgroup = subparser.add_mutually_exclusive_group()
        if bool(args.get("all")):
            for arg, (
                arg_literal,
                arg_desc,
                arg_metavar,
                arg_type,
                arg_opt,
            ) in args["all"].items():
                if arg_type in [bool, None]:
                    subparser.add_argument(
                        arg,
                        arg_literal,
                        action="store_true",
                        required=False if arg_opt == "opt" else True,
                        help=arg_desc,
                    )
                else:
                    if arg_opt in ["opt", "optional"]:
                        optional.add_argument(
                            arg,
                            arg_literal,
                            nargs=1,
                            type=arg_type,
                            help=arg_desc,
                            metavar=arg_metavar,
                        )
                    else:
                        required.add_argument(
                            arg,
                            arg_literal,
                            nargs=1,
                            type=arg_type,
                            help=arg_desc,
                            required=True,
                            metavar=arg_metavar,
                        )
        for arg, (arg_literal, arg_desc, arg_metavar, arg_type, arg_opt) in args[
            cmd
        ].items():
            if arg_type in [bool, None]:
                subgroup.add_argument(
                    arg,
                    arg_literal,
                    action="store_true",
                    required=False if arg_opt == "opt" else True,
                    help=arg_desc,
                )
            else:
                if arg_opt in ["opt", "optional"]:
                    optional.add_argument(
                        arg,
                        arg_literal,
                        nargs=1,
                        type=arg_type,
                        help=arg_desc,
                        metavar=arg_metavar,
                    )
                else:
                    required.add_argument(
                        arg,
                        arg_literal,
                        nargs=1,
                        type=arg_type,
                        help=arg_desc,
                        required=True,
                        metavar=arg_metavar,
                    )
        subparser._action_groups.append(optional)
    return parser


def cli(argv=None) -> None:
    """_summary_
    Command Line Interface entry point for user interaction.
    Args:
        argv (list, optional): Commands parsed as method input for CLI testing purposes. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    parser = setupParser(
        __cmds,
        __cmd_args,
        author=__author__,
        email=__email__,
        prog_vers=__version__,
        prog_name="dycifer",
        desc="Dynamic Integrated Circuit Performance Evaluation Tool written in Python",
    )
    subparsers = [tok for tok in __cmds.keys()] + [tok[0] for tok in __cmds.values()]
    if len(argv) == 0:  # append "help" if no arguments are given
        argv.append("-h")
    elif argv[0] in subparsers:
        if len(argv) == 1:
            argv.append("-h")  # append help when only
            # one positional argument is given
    try:
        args = parser.parse_args(argv)
        try:
            args.func(argv)
        except Exception as e:
            log.error(traceback.format_exc())
    except argparse.ArgumentError as e:  # catching unknown arguments
        log.warning(e)
    except Exception as e:
        log.error(traceback.format_exc())
    sys.exit(1)  # traceback successful command run



from curses import use_default_colors
from curses.panel import bottom_panel
import os
import pdb
from loguru import logger as log
import traceback
from pandas import DataFrame
import numpy as np
from dycifer.read import readSignals
from dycifer.utils import plotPrettyFFT
from modelling_utils import stof, timer


def mixedSignalsDynamicEval(subparser, *args, **kwargs):
    import warnings

    warnings.filterwarnings("ignore")  # supress pandas warnings
    """_summary_
    Dynamic performance evaluation of Mixed Signals circuits
    """
    argv = None
    # read data
    sysargs = args[0]
    try:
        argv = subparser.parse_args(sysargs[1:])
    except Exception as e:
        log.error(traceback.format_exc())
    # from the signals argument (containing the signals file filepath)
    # extract the signals
    signals = readSignals(argv.signals[0])
    if argv.analog_to_digital:
        sampling_freq = stof(
            argv.sampling_frequency[0]
        )  # convert the parsed string to a float
        res = argv.bit_resolution[0] if bool(argv.bit_resolution) else -1
        v_source = argv.voltage_source[0] if bool(argv.voltage_source) else 1.0
        harmonics = argv.harmonics[0] if bool(argv.harmonics) else 7
        signal_span = argv.signal_span[0] if bool(argv.signal_span) else 0.0
        noise_power = argv.noise_power[0] if bool(argv.noise_power) else -1.0
        # pdb.set_trace()
        # perform dynamic performance evaluation
        (
            spectrum,
            target_harmonics,
            signal_power,
            dc_power,
            sfdr,
            thd,
            snr,
            sndr,
            enob,
            hd2,
            hd3,
        ) = adcDynamicEval(
            signals,
            sampling_freq,
            n_bits=res,
            v_source=v_source,
            harmonics=harmonics,
            signal_span_factor=signal_span,
            asceding_bit_order=argv.ascending,
            noise_power=noise_power,
        )
        # prepare to plot resulting information
        dynamic_eval_indicators = DataFrame(
            data={
                "Signal Power (dB)": signal_power,
                "DC Power (dB)": dc_power,
                "SFDR (dB)": sfdr,
                "THD (dB)": thd,
                "SNR (dB)": snr,
                "SNDR (dB)": sndr,
                "ENOB": enob,
                "HD2 (dB)": hd2,
                "HD3 (dB)": hd3,
            },
            index=["Dynamic Evaluation Indicators"],
        )
        if argv.plot:
            plotPrettyFFT(
                spectrum[
                    spectrum.index.values >= 0
                ].index,  # plot only positive frequencies spectrum
                spectrum[spectrum.index.values >= 0]["power_db"],
                title="Signal Spectrum (dB)",
                xlabel="Frequency (GHz)",
                ylabel="Power (dB)",
                show=True,
                target_harmonics=target_harmonics,
                plot_to_terminal=argv.plot_to_terminal,
                xscale="G",
            )
        if bool(argv.output_file):
            plotPrettyFFT(
                spectrum[
                    spectrum.index >= 0
                ].index,  # plot only positive frequencies spectrum
                spectrum[spectrum.index.values >= 0]["power_db"],
                title="Signal Spectrum (dB)",
                xlabel="Frequency (GHz)",
                ylabel="Power (dB)",
                show=False,
                file_path=argv.output_file[0] + ".png",
                target_harmonics=target_harmonics,
                xscale="G",
            )
            if argv.generate_table:
                tablename = argv.output_file[0]
                dynamic_eval_indicators.to_csv(tablename + ".csv")
                dynamic_eval_indicators.to_json(tablename + ".json")
                dynamic_eval_indicators.to_markdown(tablename + ".md")
                dynamic_eval_indicators.to_latex(tablename + ".tex")
        # print indicators to console
        print()
        print(dynamic_eval_indicators.T)
    elif bool(argv.digital_to_analog):
        raise NotImplementedError("digital_to_analog: Not implemented yet.")
    elif bool(argv.sigma_delta_adc):
        raise NotImplementedError("sigma_delta_adc: Not implemented yet.")
    elif bool(argv.sigma_delta_dac):
        raise NotImplementedError("sigma_delta_dac: Not implemented yet.")
    else:
        raise ValueError(
            "No mixed-signals system class was specified was specified. Missing --analog-to-digital, --digital-to-analog, --sigma-delta-adc or --sigma-delta-dac."
        )
    print("\nPerformance evaluation finished.")
    return


@timer
def adcDynamicEval(
    signals: DataFrame,
    f_sampling: float,
    n_bits: int = -1,
    v_source: float = 1.0,
    harmonics: int = 7,
    signal_span_factor: float = 0.0,
    asceding_bit_order: bool = False,
    noise_power: float = -1.0,
) -> tuple[DataFrame, float, float, float, float, float, float, float]:
    print("\nPerforming Dynamic performance evaluation of ADC...")
    """_summary_
    Dynamic performance evaluation of Analog-to-Digital Converter circuits
    Args:
        signals (DataFrame): The signals corresponding to each of the bits generated by the ADC to be analysed.
        NOTE: signals can either be:
            - dataframe with index = time axis, and columns = digital word (in decimal, from 0 to 2^n_bits-1)
            - dataframe with index = time axis, and columns = rectangular signals corresponding to each bit of the digital word
        time_col (bool, optional): If True, the csv file is assume to have a time column. Otherwise, time column is not present.
        f_sampling (float, optional): The sampling frequency (in Hertz (Hz)) corresponding to the resolution of each signal sample. Defaults to 0.0
        v_source (float, optional): The voltage source used to feed the ADC, establishing the bottom and top voltage level decks. Defaults to 1.0 V
        n_bits (int, optional): The number of bits (resolution) of the ADC. Defaults to -1.
        harmonics (int, optional): The number of harmonics considered in the analysis of the harmonic distortion of the ADC. Defaults to 7.
        singal_span_factor (float, optional): Percentual factor determining how much of the signal's
                                                power is dispersed onto the remanescent spectrum of the signal's spectrum. Defaults to 0.0
        ascending_bit_order (bool, optional): When parsing bit signals (and not output word), indicate if the columns of each bit (in the signals dataframe)
                                                are in ascending or descending order. Defaults to False.
        noise_power (float, optional):Artificially added noise power (in dBm) to the signal. Defaults to -1.0
    Returns:
        tuple(DataFrame, float(1), float(2), float(3), float(4), float(5), float(6), float(7)):
            DataFrame: The frequency spectrum of the ADC's output signal in volt, volt squared (power) and decibels.
            float(1): Signal's power in decibels
            float(2): Signal's DC power
            float(3): Spurious Free Dynamic Range metric
            float(4): Total Harmonic Distortion metric
            float(5): Signal to Noise Ratio metric
            float(6): Signal to Noise & Distortion Ratio metric
            float(7): Effective Number of Bits (effective ADC resolution) metric
    """

    # extract the sampling frequency from the function inputs
    ts = 1.0 / f_sampling
    fs = f_sampling
    downsampling = 1
    if bool(signals.index.name):
        downsampling = int(
            round(
                (1.0 / f_sampling) / (signals.index.values[1] - signals.index.values[0])
            )
        )
        if downsampling < 1:
            raise ValueError(
                "Sampling time period must be equal or higher than the signals' time resolution."
            )
    else:
        # in case there is no time axis frame
        # automatically generate one from the sampling frequency
        t = np.arange(0, len(signals) * ts, ts)
        signals.set_index(t, inplace=True)

    # downsample the signals to the sampling frequency effectively parsed as input
    signals = signals[::downsampling]
    """
    * ***********************************************************************************
    * * If the resolution of the ADC was parsed as input, it is assumed that the signals
    * * data frame already contains the constructed dout decimal words.
    * ***********************************************************************************
    """
    if n_bits > 0 and len(signals.columns) > 1:
        raise ValueError(
            f"The number of bits was provided as input, but the signals data frame does not present the constructed digital output word of the ADC. Expected {1} signal, found {len(signals.columns)} signals."
        )
    # extract the number of bits of the ADC
    resolution = n_bits if n_bits > 0 else len(signals.columns)
    # v_step = v_source / ((2 ** resolution)-1)

    """
    * ***********************************************************************************
    * * Compute the Dout signal if no resolution is provided as input, otherwise
    * * the Dout signal is already available in the signals DataFrame in the required format
    * ***********************************************************************************
    """
    # compute the Dout signal for each row of the signals DataFrame
    dout = signals[signals.columns].copy()
    vsource = v_source
    if n_bits < 0:
        # find the average value for the signals
        means = {}
        maxis = []
        minies = []
        for col in signals.columns:
            means[col] = signals[col].mean()
            maxis.append(signals[col].max())
            minies.append(signals[col].min())

        vsource = max(maxis) + min(minies)

        for col in dout.columns:
            dout[col][dout[col] <= means[col]] = 0
            dout[col][dout[col] > means[col]] = 1

        def join_bits(row):
            if not asceding_bit_order:
                return "".join(map(lambda b: str(int(b)), row.values))
            else:
                return "".join(map(lambda b: str(int(b))), np.flipud(row.values))

        dout["bin_word"] = dout.apply(join_bits, axis=1)
        dout["dec_word"] = dout["bin_word"].apply(lambda bin: int(bin, 2))
        # recenter the decoded word in 0 and scale it to [-1; +1]
        dout["vout"] = dout["dec_word"] / (2**resolution - 1) * 2 - 1.0
        dout["vout"] = dout["vout"] * vsource
    else:
        dout["vout"] = dout[dout.columns] / (2**resolution - 1) * 2 - 1.0
        dout["vout"] = dout["vout"] * vsource
    if noise_power > 0:
        noise_watt = (10 ** (noise_power / 10)) * 1e-3
        dout["vout"] = dout["vout"] + np.random.normal(
            0, np.sqrt(noise_watt), size=len(dout)
        )
    dout = dout["vout"]
    n_samples = len(dout)
    """
    * ***********************************************************************************
    * * Fast Fourier Transform (FFT) of the Dout Signal
    * ***********************************************************************************
    """
    vout = np.abs(np.fft.fftshift(np.fft.fft(dout.values) / n_samples))  # [V]
    freq = np.fft.fftshift(np.fft.fftfreq(n_samples, ts))  # [Hz]
    power = (
        vout * vout
    )  # [V^2] - square the voltage spectrum to obtain the power spectrum
    power_db = 10 * np.log10(power)  # [dB] - convert the power spectrum to dB
    spectrum = DataFrame(
        index=freq, data={"vout": vout, "power": power, "power_db": power_db}
    )
    # positive frequencies spectrum
    pspectrum = spectrum[spectrum.index >= 0].copy()
    """
    * ***********************************************************************************
    * * Computation of :
    * * SFDR - Spurious Free Dynamic Range
    * * THD - Total Harmonic Distortion
    * * SNR - Signal to Noise Ratio
    * * SNDR - Signal to Noise and Distortion Ratio
    * * ENOB - Effective Number Of Bits
    * * INL - Integral Non-Linearity (TODO)
    * * DFL - Differential Non-Linearity (TODO)
    * ***********************************************************************************
    """
    # ********************************************
    # Obtaining the ADC's output signal power
    # ********************************************
    # determine the span of the signal's spectrum to consider it's total dispersed power
    span = np.max([1, int(np.floor(signal_span_factor * len(pspectrum.index)))])
    # obtain the signal frequency bin
    signal_bin = pspectrum["power"][
        0 + span :
    ].idxmax()  # don't count DC signal when searching for the signal bin
    # obtain the harmonics of the signal from the signal bin
    harmonic_bins = [
        pspectrum.index[
            np.abs(pspectrum.index - (mult * signal_bin))
            == np.min(np.abs(pspectrum.index - (mult * signal_bin)))
        ][0]
        for mult in range(1, harmonics + 1)
        if mult * signal_bin <= np.max(freq)
    ]
    # tones that surpass Fs are aliased back to [0, Fs/2] spectrum
    harmonic_bins = [
        pspectrum.index[
            np.abs(pspectrum.index - (fs - bin))
            == np.min(np.abs(pspectrum.index - (fs - bin)))
        ][0]
        if bin > fs / 2
        else bin
        for bin in harmonic_bins
    ]
    # indexes of the harmonic bins
    harmonic_bins_idxs = [pspectrum.index.get_loc(bin) for bin in harmonic_bins]
    harmonics_power = np.array(
        [
            np.sum(
                pspectrum["power"]
                .iloc[harmonic_bin_idx - span : harmonic_bin_idx + span]
                .values
            )
            for harmonic_bin_idx in harmonic_bins_idxs
        ]
    )
    # obtain the signal_power

    signal_power = harmonics_power[0]
    SIGNAL_POWER_DB = 10 * np.log10(signal_power)
    signal_dc_power = np.sum(pspectrum["power"].iloc[0 : 0 + span].values)
    DC_POWER_DB = 10 * np.log10(signal_dc_power)
    # ********************************************
    # Computing SFDR - Spurious Free Dynamic Range
    #  - Obtain the power of the
    #       strongest spurious component
    #       of the spectrum (excluding the
    #       DC component) and compute the SFDR.
    # ********************************************
    signal_bin_idx = harmonic_bins_idxs[0]  # get the index of the signal bin
    spurious_spectrum = pspectrum["power"].copy()
    # erase the signal bin from the spurious spectrum
    spurious_spectrum.iloc[signal_bin_idx - span : signal_bin_idx + span] = np.min(
        pspectrum["power"]
    )
    # erase the signal's DC component from the spurious spectrum
    spurious_spectrum.iloc[0 : 0 + span] = np.min(pspectrum["power"])
    # find the strongest spurious component
    spur_bin = spurious_spectrum.idxmax()
    spur_bin_idx = pspectrum.index.get_loc(spur_bin)
    # measure the power of the strongest spurious component
    spur_power = np.sum(
        spurious_spectrum.iloc[spur_bin_idx - span : spur_bin_idx + span].values
    )
    # compute the SFDR
    SFDR = 10 * np.log10(signal_power / spur_power)
    # ********************************************
    # Computing THD - Total Harmonic Distortion
    #  - Obtain the power of each harmonic component
    # ********************************************
    # compute the total power of the sum of the harmonics
    THD = np.nan
    SNR = np.nan
    SNDR = np.nan
    ENOB = np.nan
    HD2 = np.nan
    HD3 = np.nan
    try:
        total_distortion_power = np.sum(harmonics_power[1:])
        THD = 10 * np.log10(total_distortion_power / harmonics_power[0])
        # ********************************************
        # Computing SNR - Signal to Noise Ratio
        #  - Obtain the noise power in the spectrum
        # ********************************************
        noise_power = (
            np.sum(pspectrum["power"].values)
            - signal_dc_power
            - signal_power
            - total_distortion_power
        )
        SNR = 10 * np.log10(signal_power / noise_power)
        # ********************************************
        # Computing SNDR - Signal to Noise & Distortion Ratio
        #  - Add the noise and distortion power in
        #     the spectrum and compare them to the
        #     signal power
        # ********************************************
        SNDR = 10 * np.log10(signal_power / (noise_power + total_distortion_power))
        # ********************************************
        # Computing ENOB - Effective Number of Bits
        #  - Check deterioration of the ADC's
        #    ideal resolution because of the SNDR
        # ********************************************
        ENOB = (SNDR - 1.76) / 6.02
        # ********************************************
        # Computing HD2 and HD3 - Fractional Harmonic
        # Distortion of Second and Third order
        # harmonics
        # ********************************************
        HD2 = 10 * np.log10(harmonics_power[1] / harmonics_power[0])
        HD3 = 10 * np.log10(harmonics_power[2] / harmonics_power[0])
    except IndexError:
        log.warning(
            f"\nTried to access an harmonic that was not found in the frequency domain of the analysed spectrum.\nThis is likely due to the fact that the chosen sampling frequency is \nviolating the Nyquist Theorem in relation to the 2nd or 3rd Order Harmonics.\nTry to increase the sampling frequency."
        )
    target_harmonics = list(zip(harmonic_bins, 10 * np.log10(harmonics_power)))
    return (
        spectrum,
        target_harmonics,
        SIGNAL_POWER_DB,
        DC_POWER_DB,
        SFDR,
        THD,
        SNR,
        SNDR,
        ENOB,
        HD2,
        HD3,
    )


def dacDynamicEval(subparser, *args, **kwargs):
    """_summary_
    Dynamic performance evaluation of Digital-to-Analog Converter circuits
    """
    raise NotImplementedError("dacDynamicEval: Not implemented yet.")


def sigmaDeltaAdcDynamicEval(*args, **kwargs):
    """_summary_
    Dynamic evaluation of Sigma Delta Analog-to-Digital Converter circuits
    Raises:
        NotImplementedError: _description_
    """
    raise NotImplementedError("sigmaDeltaAdcDynamicEval: Not implemented yet")


def sigmaDeltaDacDynamicEval(*args, **kwargs):
    """_summary_
    Dynamic evaluation of Sigma Delta Analog-to-Digital Converter circuits
    Raises:
        NotImplementedError: _description_
    """
    raise NotImplementedError("sigmaDeltaAdcDynamicEval: Not implemented yet")



import os
from modelling_utils import Scale


def getParent(path, levels=0):
    """_summary_
    Get the parent directory of a path
    according to the specified level of
    depth in the tree
    Args:
        path    (str)   : child path of the parent directory
        levels  (int)   : number of levels to go up in the tree
    """
    common = path
    for _ in range(levels + 1):
        common = os.path.dirname(os.path.abspath(common))
    return common


def plotPrettyFFT(
    freq,
    power,
    title: str = "",
    xlabel: str = "",
    ylabel: str = "",
    file_path: str = None,
    xlog: bool = False,
    show: bool = False,
    target_harmonics: list = None,
    plot_to_terminal: bool = False,
    xscale: str = "",
    yscale: str = "",
    xunit: str = "Hz",
    yunit: str = "dB",
):
    """_summary_
    Plot a pretty FFT plot
    Args:
        freq (array/list): frequency array
        power (array/list): power array
        title (str)     : title of the plot
        xlabel (str)    : x-axis label
        ylabel (str)    : y-axis label
        file_path (str) : path to save the plot
        xlog (bool)     : log scale
        show (bool)     : show the plot
        target_harmonics (list): list (frequency, power) tuples of target harmonics to highlight
        xscale (str)    : x-axis scale
        yscale (str)    : y-axis scale
        xunit (str)     : x-axis unit. Default is Hz (hertz)
        yunit (str)     : y-axis unit. Default is dB (decibel)
    NOTE:
        kwargs example:
        kwargs = {
            "linefmt":"b-",
            "markerfmt":"bD",
            "basefmt":"r-",
        }
    """
    from loguru import logger as log
    import plotext as tplt
    import matplotlib.pyplot as plt
    from matplotlib import rcParams
    from numpy import min, where, max, floor, abs, array, append, sort, argsort
    import pdb
    from itertools import cycle

    plt.rc("axes", titlesize=14)  # fontsize of the axes title
    plt.rc("axes", labelsize=12)  # fontsize of the x and y labels
    plt.rc("xtick", labelsize=12)  # fontsize of the tick labels
    plt.rc("ytick", labelsize=12)  # fontsize of the tick labels
    plt.rc("legend", fontsize=11)  # legend fontsize
    plt.rc("font", size=11)  # controls default text sizes
    # define font family to use for all text
    rcParams["font.family"] = "serif"
    # plt.figure(figsize=(8,6))
    # markerline, stemlines, baseline = plt.stem(freq, power, bottom=min(power), use_line_collection=True, linefmt="b-", markerfmt="bD", basefmt="r-")
    # markerline.set_markerfacecolor("none")
    # stemlines.set_color("black")
    x_scale = 1.0
    if xscale != "":
        letter_scales = [letter for (letter, _) in [elem.value for elem in Scale]]
        value_scales = [value for (_, value) in [elem.value for elem in Scale]]
        if xscale in letter_scales:
            x_scale = value_scales[letter_scales.index(xscale)]
        else:
            log.warning(
                f"\n[xscale] {xscale} not recognized. Possible scales are: {letter_scales}"
            )
    y_scale = 1.0
    if yscale != "":
        letter_scales = [letter for letter, _ in Scale.values]
        value_scales = [value for _, value in Scale.values]
        if yscale in letter_scales:
            y_scale = value_scales[letter_scales.index(yscale)]
        else:
            log.warning(
                f"\n[yscale] {yscale} not recognized. Possible scales are: {letter_scales}"
            )

    harmonics = [
        (freq[where(power == max(power[1:]))][0], max(power[1:]))
    ]  # don't count the dc component
    harmonics = target_harmonics if bool(target_harmonics) else harmonics
    # add the target harmonics to the plot in case they were folded into the positive frequency axis
    new_freq = append(freq, [harmonics[i][0] for i in range(len(harmonics))])
    freq = sort(new_freq)
    power = append(power, [harmonics[i][1] for i in range(len(harmonics))])[
        argsort(new_freq)
    ]
    if plot_to_terminal and show:
        tplt.plot(freq, power)
        tplt.ylim(min(power), max(power) + abs(max(power)) * 1.1)
        # plot the marks of the harmonics
        colours = cycle(
            ["red+", "gray+", "green+", "white", "magenta+", "cyan+", "yellow+"]
        )
        # line_styles = cycle(["-.", "--", "-.", "-", "--"])
        labels = [
            f"H{x}@{f:.3f} {xscale}{xunit}"
            for (x, f) in enumerate([f for f, _ in harmonics], start=1)
        ]
        """
        tplt.horizontal_line(
            max(power[1:]),
            color="white",
        )
        """
        for color, harmonic, label in zip(colours, harmonics, labels):
            x = harmonic[0]
            y = harmonic[1]
            tplt.plot([x], [y], marker="sd", color=color)
            tplt.text(label, x, y - 0.06 * abs(y), color=color)

        tplt.grid()
        tplt.title(title + f"- H1 Power={max(power[1:]):.2f} ({yscale}{yunit})")
        tplt.xlabel(xlabel)
        tplt.ylabel(ylabel)
        if xlog:
            tplt.xscale("log")
        tplt.theme("dark")
        tplt.show()
        # clear data
        if bool(file_path):
            if not os.path.exists(getParent(file_path)):
                raise FileNotFoundError(
                    f"Directory does not exist : {getParent(file_path)}"
                )
            tplt.savefig(file_path)
        tplt.cld()
    else:
        plt.plot(freq / x_scale, power / y_scale, "b-")
        markerline, _, _ = plt.stem(
            freq / x_scale,
            power / y_scale,
            bottom=min(power / y_scale),
            use_line_collection=True,
            linefmt="b-",
            markerfmt="none",
            basefmt="r-",
        )
        markerline.set_markerfacecolor("none")
        colours = cycle(["red", "brown", "green", "black", "magenta", "cyan", "yellow"])
        # line_styles = cycle(["-.", "--", "-.", "-", "--"])
        labels = [
            f"H{x}@{f/x_scale:.3f} ({xscale}{xunit})"
            for (x, f) in enumerate([f for f, _ in harmonics], start=1)
        ]
        plt.axhline(
            max(power[1:] / y_scale),
            color="k",
            linestyle="--",
            linewidth=3,
            label=f"H1 Power={max(power[1:]/y_scale):.2f} ({yscale}{yunit})",
        )
        for color, harmonic, label in zip(colours, harmonics, labels):
            x = harmonic[0]
            y = harmonic[1]
            plt.plot(x / x_scale, [y / y_scale], "D", color=color, label=label)

        plt.grid()
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.legend(loc="upper right")
        if xlog:
            plt.xscale("log")
        if show:
            plt.show()
        if bool(file_path):
            if not os.path.exists(getParent(file_path)):
                raise FileNotFoundError(
                    f"Directory does not exist : {getParent(file_path)}"
                )
            plt.savefig(file_path)
    plt.clf()



import os
import traceback as tb
from modelling_utils import (
    Scale,
    Units,
    stof,
)
from pandas import (
    read_csv,
    DataFrame,
)


def readSignals(file_path: str = None) -> DataFrame:
    """_summary_
    Reads a time series table data into a pandas DataFrame and returns it.
    Args:
        file_path (str, optional): The file path of the Comma Separated Values (.CSV) file generated by Cadence (or other EDA software).
                                    Defaults to None.
        time_col (bool, optional): If True, the csv file is assume to have a time column. Otherwise, time column is discarded.
    Returns:
        DataFrame: the pandas DataFrame containing the time series data with all the correspondant signals.
    """
    time_col = "t"
    if file_path is None:
        raise ValueError("The file path was not provided.")
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File {file_path} not found.")
    signals = None
    try:
        signals = read_csv(file_path, header=0)
    except Exception as e:
        tb.format_exc()
        return None
    # preprocess the headers of the data - search for the time column
    possible_time_cols = ["time", "t", "Time", "T", "TIME"]
    ordered = False
    signals_columns_order = signals.columns.tolist()
    for column in list(signals.columns):
        tokens = column.split(" ")
        for token in tokens:
            if token in possible_time_cols + [
                Units.TIME.value,
                f"[{Units.TIME.value}]",
                f"({Units.TIME.value})",
            ]:
                # place the time column in the first column of the DataFrame
                signals_columns_order = [column] + list(
                    signals.drop(columns=[column]).columns
                )
                ordered = True
                break
        if ordered:
            break
    if ordered:
        return signals[signals_columns_order].set_index(signals_columns_order[0])
    else:
        return signals



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dycifer']

package_data = \
{'': ['*']}

install_requires = \
['Markdown>=3.3.7,<4.0.0',
 'argparse>=1.4.0,<2.0.0',
 'colorama>=0.4.5,<0.5.0',
 'loguru>=0.6.0,<0.7.0',
 'matplotlib>=3.5.2,<4.0.0',
 'modelling-utils>=0.2.11,<0.3.0',
 'numpy>=1.22.4,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'plotext>=5.0.2,<6.0.0',
 'pyfiglet>=0.8.post1,<0.9',
 'tabulate>=0.8.9,<0.9.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['dycifer = dycifer.dycifer:cli']}

setup_kwargs = {
    'name': 'dycifer',
    'version': '0.1.0',
    'description': 'DYCIFER (Dynamic Circuits Performance Evaluation Tool) is a tool dedicated to the calculation of the performance indicators of integrated circuits operating at relatively high frequencies.',
    'long_description': None,
    'author': 'Diogo André S. Dias',
    'author_email': 'das.dias6@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/das-dias/dyciferpy',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
